@@ -3,7 +3,7 @@
 /*
 
     This file is part of the extensible drawing editor Ipe.
-    Copyright (c) 1993-2019 Otfried Cheong
+    Copyright (c) 1993-2020 Otfried Cheong
 
     Ipe is free software; you can redistribute it and/or modify it
     under the terms of the GNU General Public License as published by
@@ -33,6 +33,11 @@
 
 #include "ipelua.h"
 
+#include "ipethumbs.h"
+
+#include <cairo.h>
+#include <cairo-quartz.h>
+
 #include <cstdio>
 #include <cstdlib>
 
@@ -44,13 +49,33 @@ using namespace ipelua;
 // in ipeui_cocoa:
 extern NSImage *colorIcon(double red, double green, double blue, int pixels);
 
+// in ipebitmap_unix:
+extern bool cgImageDecode(CGImageRef bitmap, Buffer pixelData);
+
 // --------------------------------------------------------------------
 
 static const char * const snapbutton_action[] = {
   "snapvtx", "snapctl", "snapbd", "snapint",
-  "snapgrid", "snapangle", "snapauto",
+  "snapgrid", "snapangle", "snapcustom", "snapauto",
   "grid_visible" };
 
+static const char * const touchbar_action[] = {
+  "escape", "set_origin", "set_direction", "set_line",
+  "show_axes", "reset_direction", "set_tangent_direction",
+  "snapvtx", "snapctl", "snapbd", "snapint",
+  "snapgrid", "snapangle", "snapcustom", "snapauto" };
+
+#if MAC_OS_X_VERSION_MAX_ALLOWED >= 101202
+static const char * const touchbar_titles[] = {
+  "esc", "org", "dir", "line", "axes", "reset", "tangent",
+  nullptr, nullptr, nullptr, nullptr, nullptr, nullptr, nullptr, nullptr };
+
+static const char * const touchbar_customization_label[] = {
+  "Escape", "Set origin", "Set direction", "Set to edge", "Show axes", "Reset direction", "Set to tangent direction",
+  "Snap to vertex", "Snap to control point", "Snap to boundary", "Snap to intersection", "Snap to grid",
+  "Angular snap", "Snap to custom grid", "Automatic snap" };
+#endif
+
 @interface IpeAction : NSObject
 @property NSString *name;
 @property NSString *title;
@@ -108,10 +133,10 @@ static const char * const snapbutton_action[] = {
       self.action = @selector(delete:);
     else if ([_ipeAction isEqualToString:@"select_all"])
       self.action = @selector(selectAll:);
-    else if ([_ipeAction isEqualToString:@"undo"])
-      self.action = @selector(undo:);
     else if ([_ipeAction isEqualToString:@"fullscreen"])
       self.action = @selector(toggleFullScreen:);
+    else if ([_ipeAction isEqualToString:@"recent_file"])
+      self.action = @selector(ipeRecentFileAction:);
     // make sure these work when no window exists
     else if ([@[@"show_configuration", @"new_window", @"open",
 		@"manual", @"about_ipelets"]
@@ -125,7 +150,7 @@ static const char * const snapbutton_action[] = {
 
 // --------------------------------------------------------------------
 
-static NSImage *colorIcon(Color color, int pixels = 24)
+static NSImage *colorIcon(Color color, int pixels = 20)
 {
   double red = color.iRed.toDouble();
   double green = color.iGreen.toDouble();
@@ -133,32 +158,16 @@ static NSImage *colorIcon(Color color, int pixels = 24)
   return colorIcon(red, green, blue, pixels);
 }
 
-static NSImage *loadIcon(String action)
-{
-  String dir = ipeIconDirectory();
-  String fname = dir + action + ".tiff";
-  if (!Platform::fileExists(fname)) {
-    fname = dir + action + ".png";
-    if (!Platform::fileExists(fname))
-      return nil;
-  }
-  return [[NSImage alloc] initWithContentsOfFile:I2N(fname)];
-}
-
 // --------------------------------------------------------------------
 
-@interface IpeWindowDelegate :
-  NSObject <NSWindowDelegate, NSToolbarDelegate, IpeControlsDelegate>
+#if MAC_OS_X_VERSION_MAX_ALLOWED < 101202
+@protocol NSTouchBarDelegate
+@end
+#endif
 
-- (instancetype) initFor:(AppUi *) ui;
-- (void) ipeMenuAction:(id) sender;
-- (void) ipeToolbarAction:(id) sender;
-- (void) ipeAbsoluteButton:(id) sender;
-- (void) ipeSnapButton:(id) sender;
-- (void) ipeSelectorChanged:(id) sender;
-- (void) toggleSnapbarShown:(id) sender;
-- (void) ipeSubmenu:(id) sender;
-- (BOOL) ipeIsModified:(id) sender;
+// Window delegate is also an NSWindowController object to handle touch bar
+@interface IpeWindowDelegate : NSWindowController
+  <NSWindowDelegate, NSToolbarDelegate, IpeControlsDelegate, NSTouchBarDelegate>
 @end
 
 @implementation IpeWindowDelegate {
@@ -167,7 +176,7 @@ static NSImage *loadIcon(String action)
 
 - (instancetype) initFor:(AppUi *) ui
 {
-  self = [super init];
+  self = [super initWithWindow:nil];
   if (self) {
     appui = ui;
   }
@@ -187,6 +196,8 @@ static NSImage *loadIcon(String action)
   }
   if (item.action == @selector(toggleSnapbarShown:))
     return appui->validateMenuItem(item, @"snapbar");
+  if (item.action == @selector(togglePropertiesShown:))
+    return appui->validateMenuItem(item, @"propertiespanel");
   if (item.tag >= AppUi::ESubmenuGridSize && item.tag < AppUi::ESubmenuFin)
     appui->fillDynamicSubmenu(item);
   return YES;
@@ -272,6 +283,11 @@ static NSImage *loadIcon(String action)
   appui->toggleSnapbarShown();
 }
 
+- (void) togglePropertiesShown:(id) sender
+{
+  appui->togglePropertiesShown();
+}
+
 // --------------------------------------------------------------------
 
 - (void) pathViewAttributeChanged:(String) attr
@@ -314,13 +330,15 @@ static NSImage *loadIcon(String action)
 {
   NSToolbarItem *t =
     [[NSToolbarItem alloc] initWithItemIdentifier:itemIdentifier];
-  t.image = loadIcon(N2I(itemIdentifier));
+  t.image = appui->loadIcon(N2I(itemIdentifier), false, 32);
   if (![itemIdentifier hasPrefix:@"mode_"])
     t.visibilityPriority = NSToolbarItemVisibilityPriorityLow;
   t.action = @selector(ipeToolbarAction:);
   IpeAction *s = appui->findAction(itemIdentifier);
-  if (s)
+  if (s) {
     t.toolTip = s.title;
+    t.label = s.title;
+  }
   return t;
 }
 
@@ -331,7 +349,7 @@ static NSImage *loadIcon(String action)
 	     @"fit_width", @"keyboard",
 	     @"mode_select", @"mode_translate", @"mode_rotate",
 	     @"mode_stretch", @"mode_shear", @"mode_graph",
-	     @"mode_pan", @"mode_shredder",
+	     @"mode_pan", @"mode_shredder", @"mode_laser",
 	     @"mode_label", @"mode_math", @"mode_paragraph",
 	     @"mode_marks", @"mode_rectangles1", @"mode_rectangles2",
 	     @"mode_rectangles3", @"mode_parallelogram", @"mode_lines",
@@ -348,7 +366,7 @@ static NSImage *loadIcon(String action)
 {
   return @[ @"mode_select", @"mode_translate", @"mode_rotate",
 	     @"mode_stretch", @"mode_shear", @"mode_graph",
-	     @"mode_pan", @"mode_shredder",
+	     @"mode_pan", @"mode_shredder", @"mode_laser",
 	     @"mode_label", @"mode_math", @"mode_paragraph",
 	     @"mode_marks", @"mode_rectangles1", @"mode_rectangles2",
 	     @"mode_rectangles3", @"mode_parallelogram", @"mode_lines",
@@ -363,7 +381,7 @@ static NSImage *loadIcon(String action)
 {
   return @[ @"mode_select", @"mode_translate", @"mode_rotate",
 	     @"mode_stretch", @"mode_shear", @"mode_graph",
-	     @"mode_pan", @"mode_shredder",
+	     @"mode_pan", @"mode_shredder", @"mode_laser",
 	     @"mode_label", @"mode_math", @"mode_paragraph",
 	     @"mode_marks", @"mode_rectangles1", @"mode_rectangles2",
 	     @"mode_rectangles2", @"mode_rectangles3",
@@ -373,6 +391,58 @@ static NSImage *loadIcon(String action)
 	     @"mode_circle1", @"mode_circle2", @"mode_circle3",
 	     @"mode_ink" ];
 }
+
+#if MAC_OS_X_VERSION_MAX_ALLOWED >= 101202
+#pragma clang diagnostic push
+#pragma clang diagnostic ignored "-Wdeprecated-declarations"
+- (NSTouchBar *) makeTouchBar
+{
+  NSTouchBar *bar = [[NSTouchBar alloc] init];
+  bar.delegate = self;
+  bar.defaultItemIdentifiers = @[ @"show_axes", @"set_origin", @"set_direction", @"snapvtx", @"snapctl", @"snapbd", @"snapint", @"snapgrid" ];
+  bar.customizationIdentifier = @"org.otfried.ipe.mainbar7_2_11";
+  bar.customizationAllowedItemIdentifiers =
+    @[ @"show_axes", @"set_origin", @"set_direction", @"reset_direction",
+	@"set_line", @"set_tangent_direction",
+	@"snapvtx", @"snapctl", @"snapbd", @"snapint", @"snapgrid", @"snapangle", @"snapcustom", @"snapauto" ];
+  bar.escapeKeyReplacementItemIdentifier = @"escape";
+  return bar;
+}
+
+- (NSTouchBarItem *) touchBar: (NSTouchBar *) bar
+	makeItemForIdentifier: (NSTouchBarItemIdentifier) identifier
+{
+  String idf = N2I(identifier);
+  for (size_t tag = 0; tag < (sizeof(touchbar_action)/sizeof(const char *)); ++tag) {
+    if (idf == touchbar_action[tag]) {
+      NSCustomTouchBarItem *item = [[NSCustomTouchBarItem alloc] initWithIdentifier:identifier];
+      auto title = touchbar_titles[tag];
+      NSButton *b;
+      if (title == nullptr) {
+	auto image = appui->loadIcon(touchbar_action[tag], true);
+	b = [NSButton buttonWithTitle:identifier image:image target:self action:@selector(ipeTouchBar:)];
+	b.buttonType = NSPushOnPushOffButton;
+	b.imagePosition = NSImageOnly;
+      } else {
+	b = [NSButton buttonWithTitle:C2N(title) target:self action:@selector(ipeTouchBar:)];
+	if (idf == "show_axes")
+	  b.buttonType = NSPushOnPushOffButton;
+      }
+      b.tag = tag;
+      item.view = b;
+      item.customizationLabel = C2N(touchbar_customization_label[tag]);
+      return item;
+    }
+  }
+  return nil;
+}
+#pragma clang diagnostic pop
+#endif
+
+- (void) ipeTouchBar:(id) sender
+{
+  appui->action(touchbar_action[[sender tag]]);
+}
 @end
 
 // --------------------------------------------------------------------
@@ -386,7 +456,7 @@ void AppUi::addRootMenu(int id, const char *name)
 
 void AppUi::addItem(NSMenu *menu, const char *title, const char *name)
 {
-  if (title == 0) {
+  if (title == nullptr) {
     if (build_menus)
       [menu addItem:[NSMenuItem separatorItem]];
   } else {
@@ -494,7 +564,7 @@ void AppUi::addItem(int id, const char *title, const char *name)
   addItem(menu, title, name);
 }
 
-static NSMenu * __unsafe_unretained current_submenu = 0;
+static NSMenu * __unsafe_unretained current_submenu = nullptr;
 
 void AppUi::startSubMenu(int id, const char *name, int tag)
 {
@@ -505,7 +575,7 @@ void AppUi::startSubMenu(int id, const char *name, int tag)
 
   NSString *title = ipeui_set_mnemonic(C2N(name), nil);
   NSMenuItem *item = [[NSMenuItem alloc] initWithTitle:title
-						action:NULL
+						action:nullptr
 					 keyEquivalent:@""];
   NSMenu *submenu = [[NSMenu alloc] initWithTitle:title];
   item.submenu = submenu;
@@ -570,6 +640,8 @@ AppUi::AppUi(lua_State *L0, int model)
   [tb setSizeMode:NSToolbarSizeModeSmall];
   [iWindow setToolbar:tb];
 
+  [iWindow setWindowController:iDelegate];
+
   iContent = [[NSView alloc] initWithFrame:contentRect];
 
   makePropertiesTool();
@@ -621,11 +693,9 @@ AppUi::AppUi(lua_State *L0, int model)
   layout(iSnapBar, iView, "l=l");
   layout(iPropertiesBox, iContent, "l=l");
   layout(iPropertiesBox, iContent, "t=t");
-  layout(iPropertiesBox, iView, "r=l");
   layout(iView, iContent, "r=r");
   layout(iView, iStatus, "b=t");
   layout(iLayerBox, iContent, "l=l");
-  layout(iLayerBox, iView, "r=l");
   layout(iLayerBox, iPropertiesBox, "t=b");
   layout(iLayerBox, iStatus, "b=t");
   layout(iStatus, iContent, "l=l");
@@ -635,9 +705,23 @@ AppUi::AppUi(lua_State *L0, int model)
   layout(iMouseIndicator, iZoomIndicator, "r=l");
   layout(iZoomIndicator, iContent, "r=r");
 
+  lua_getglobal(L, "prefs");
+  lua_getfield(L, -1, "osx_properties_width");
+  if (lua_isnumber(L, -1)) {
+    double width = lua_tonumber(L, -1);
+    layout(iLayerBox, nil, "w>0", width);
+  }
+  lua_pop(L, 2);
+
   iViewToSnapBar = layout(iView, iSnapBar, "t=b", 0.0, 1.0, NO);
   iViewToTop = layout(iView, iContent, "t=t", 0.0, 1.0, NO);
-  activateConstraint(iViewToSnapBar, YES);
+  iViewToSnapBar.active = YES;
+
+  iViewToProperties = layout(iPropertiesBox, iView, "r=l", 0.0, 1.0, NO);
+  iViewToLayers = layout(iLayerBox, iView, "r=l", 0.0, 1.0, NO);
+  iViewToLeft = layout(iView, iContent, "l=l", 0.0, 1.0, NO);
+  iViewToProperties.active = YES;
+  iViewToLayers.active = YES;
 
   [iStatus
     setContentHuggingPriority:NSLayoutPriorityDefaultLow
@@ -670,14 +754,16 @@ void AppUi::makePropertiesTool()
   iPropertiesBox = [[NSBox alloc] initWithFrame:NSMakeRect(0.,0.,50.,300.)];
   [iPropertiesBox setTitle:@"Properties"];
 
-  for (int i = 0; i <= EUiSymbolSize; ++i) {
-    iButton[i] = [[NSButton alloc] initWithFrame:NSMakeRect(0., 0., 20., 20.)];
-    [iButton[i] setButtonType:NSMomentaryPushInButton];
-    [iButton[i] setImagePosition:NSImageOnly];
-    [iButton[i] setBezelStyle:NSRegularSquareBezelStyle];
-    iButton[i].action = @selector(ipeAbsoluteButton:);
-    iButton[i].tag = i;
-    addToLayout(iPropertiesBox, iButton[i]);
+  for (int i = 0; i <= EUiOpacity; ++i) {
+    if (i != EUiDashStyle && i != EUiMarkShape && i != EUiOpacity) {
+      iButton[i] = [[NSButton alloc] initWithFrame:NSMakeRect(0., 0., 12., 12.)];
+      [iButton[i] setButtonType:NSMomentaryPushInButton];
+      [iButton[i] setImagePosition:NSImageOnly];
+      [iButton[i] setBezelStyle:NSRegularSquareBezelStyle];
+      iButton[i].action = @selector(ipeAbsoluteButton:);
+      iButton[i].tag = i;
+      addToLayout(iPropertiesBox, iButton[i]);
+    }
 
     iSelector[i] = [[NSPopUpButton alloc]
 		     initWithFrame:NSMakeRect(0., 0., 50., 20.)
@@ -687,10 +773,7 @@ void AppUi::makePropertiesTool()
     iSelector[i].tag = i;
     addToLayout(iPropertiesBox, iSelector[i]);
   }
-  iButton[EUiMarkShape].enabled = NO;
-  // [iButton[EUiMarkShape] setImageDimsWhenDisabled:NO];
-
-  iModeIndicator = [[NSImageView alloc] initWithFrame:NSMakeRect(0, 0, 20, 20)];
+  iModeIndicator = [[NSImageView alloc] initWithFrame:NSMakeRect(0., 0., 12., 12.)];
   iModeIndicator.editable = NO;
   iModeIndicator.image = loadIcon("mode_select");
   addToLayout(iPropertiesBox, iModeIndicator);
@@ -699,7 +782,6 @@ void AppUi::makePropertiesTool()
   [iButton[EUiFill] setImage:colorIcon(Color(1000, 1000, 0))];
   [iButton[EUiPen] setImage:loadIcon("pen")];
   [iButton[EUiTextSize] setImage:loadIcon("mode_label")];
-  [iButton[EUiMarkShape] setImage:loadIcon("mode_marks")];
   [iButton[EUiSymbolSize] setImage:loadIcon("mode_marks")];
 
   [iButton[EUiStroke] setToolTip:@"Absolute stroke color"];
@@ -714,6 +796,8 @@ void AppUi::makePropertiesTool()
   [iSelector[EUiTextSize] setToolTip:@"Symbolic text size"];
   [iSelector[EUiMarkShape] setToolTip:@"Mark shape"];
   [iSelector[EUiSymbolSize] setToolTip:@"Symbolic symbol size"];
+  [iSelector[EUiDashStyle] setToolTip:@"Dash style"];
+  [iSelector[EUiOpacity] setToolTip:@"Opacity"];
 
   iViewNumber = [[NSButton alloc] initWithFrame:NSMakeRect(0, 0, 10, 20)];
   [iViewNumber setButtonType:NSMomentaryPushInButton];
@@ -761,28 +845,50 @@ void AppUi::makePropertiesTool()
 
   NSView *inside = [iPropertiesBox contentView];
   id guide = layoutGuide(iPropertiesBox);
-  layout(iButton[EUiStroke], inside, "t=t", 2.0);
-  for (int i = 0; i <= EUiSymbolSize; ++i) {
-    layout(iButton[i], iButton[i], "w=h");
-    layout(iButton[i], inside, "l=l");
-    layout(iButton[i], guide, "r=l");
-    layout(guide, iSelector[i], "r=l");
-    layout(iSelector[i], inside, "r=r");
-    layout(iSelector[i], iButton[i], "y=y");
-    if (i != EUiStroke && i != EUiTextSize)
-      layout(iButton[i], iButton[i-1], "t=b", 2.0);
-  }
   layout(guide, inside, "t=t");
   layout(guide, inside, "b=b");
-  layout(iPathView, iButton[EUiPen], "t=b", 2.0);
-  layout(iButton[EUiTextSize], iPathView, "t=b", 2.0);
-  layout(iPathView, guide, "l=r");
-  layout(iPathView, inside, "r=r");
-  layout(iModeIndicator, iPathView, "t=t");
-  layout(iModeIndicator, iPathView, "b=b");
+
+  // left-right layout
+  for (int i = 0; i <= EUiOpacity; ++i) {
+    if (i != EUiDashStyle && i != EUiMarkShape && i != EUiOpacity) {
+      layout(iButton[i], inside, "l=l");
+      layout(iButton[i], guide, "r=l");
+      layout(iButton[i], nil, "h>0", 28.);
+      layout(iButton[i], nil, "w=0", 30.);
+    }
+    layout(guide, iSelector[i], "r=l");
+    layout(iSelector[i], inside, "r=r");
+  }
   layout(iModeIndicator, inside, "l=l");
   layout(iModeIndicator, guide, "r=l");
-  layout(iViewNumber, iButton[EUiSymbolSize], "t=b", 2.0);
+  layout(iPathView, guide, "l=r");
+  layout(iPathView, inside, "r=r");
+
+  // top-down layout
+  layout(iButton[EUiStroke], inside, "t=t", 2.0);
+  layout(iSelector[EUiStroke], iButton[EUiStroke], "t=t");
+  layout(iButton[EUiFill], iButton[EUiStroke], "t=b", 2.0);
+  layout(iSelector[EUiFill], iButton[EUiFill], "y=y");
+  layout(iButton[EUiPen], iButton[EUiFill], "t=b", 2.0);
+  layout(iSelector[EUiPen], iButton[EUiPen], "t=t");
+  layout(iSelector[EUiDashStyle], iSelector[EUiPen], "t=b", 2.0);
+  layout(iSelector[EUiDashStyle], iButton[EUiPen], "b=b");
+
+  layout(iPathView, iButton[EUiPen], "t=b", 6.0);
+  layout(iModeIndicator, iPathView, "t=t");
+  layout(iModeIndicator, iPathView, "b=b");
+
+  layout(iButton[EUiTextSize], iPathView, "t=b", 2.0);
+  layout(iSelector[EUiTextSize], iButton[EUiTextSize], "y=y");
+
+  layout(iButton[EUiSymbolSize], iButton[EUiTextSize], "t=b", 2.0);
+  layout(iSelector[EUiMarkShape], iButton[EUiSymbolSize], "t=t");
+  layout(iSelector[EUiSymbolSize], iSelector[EUiMarkShape], "t=b", 2.0);
+  layout(iSelector[EUiSymbolSize], iButton[EUiSymbolSize], "b=b");
+
+  layout(iSelector[EUiOpacity], iButton[EUiSymbolSize], "t=b", 2.0);
+
+  layout(iViewNumber, iSelector[EUiOpacity], "t=b", 2.0);
   layout(iViewNumber, inside, "l=l");
   layout(iPageNumber, iViewNumber, "l=r", 2.0);
   layout(iPageNumber, iViewNumber, "t=t");
@@ -795,9 +901,9 @@ void AppUi::makePropertiesTool()
   layout(inside, iViewMarked, "b=b", 5.0);
 
   layout(guide, nil, "w=0", 5.0);
-  layout(iPathView, iPathView, "h=w", 0.0, 0.3);
+  layout(iPathView, nil, "h>0", 40.0);
+  //layout(iPathView, iPathView, "h=w", 0.0, 0.3);
   layout(iViewNumber, iPageNumber, "w=w");
-  layout(iModeIndicator, iModeIndicator, "w=h");
 }
 
 void AppUi::makeSnapBar()
@@ -807,7 +913,7 @@ void AppUi::makeSnapBar()
 
   iSnapBar = [[NSView alloc] initWithFrame:NSMakeRect(0., 0., 600., 32.)];
 
-  for (int i = 0; i < 8; ++i) {
+  for (int i = 0; i < NUMSNAPBUTTONS; ++i) {
     iSnapButton[i] = [[NSButton alloc] initWithFrame:NSMakeRect(0, 0, 32, 32)];
     iSnapButton[i].buttonType = NSPushOnPushOffButton;
     iSnapButton[i].imagePosition = NSImageOnly;
@@ -839,7 +945,8 @@ void AppUi::makeSnapBar()
   iSelector[EUiAngleSize].tag = EUiAngleSize;
   addToLayout(iSnapBar, iSelector[EUiAngleSize]);
 
-  for (int i = 0; i < 8; ++i) {
+  // Order: "vtx", "ctl", "bd", "int", "grid", "angle", "custom", "auto", "visible"
+  for (int i = 0; i < NUMSNAPBUTTONS; ++i) {
     layout(iSnapButton[i], iSnapBar, "t=t", MARGIN);
     layout(iSnapBar, iSnapButton[i], "b=b", MARGIN);
     if (0 < i && i < 5)
@@ -857,8 +964,9 @@ void AppUi::makeSnapBar()
   layout(iSnapButton[5], iSelector[EUiGridSize], "l=r", PAD);
   layout(iSelector[EUiAngleSize], iSnapButton[5], "l=r", PAD);
   layout(iSnapButton[6], iSelector[EUiAngleSize], "l=r", PAD);
-  layout(iSnapButton[7], iSnapButton[6], "l>r", PAD);
-  layout(iSnapBar, iSnapButton[7], "r=r", MARGIN);
+  layout(iSnapButton[7], iSnapButton[6], "l=r", PAD);
+  layout(iSnapButton[8], iSnapButton[7], "l>r", PAD);
+  layout(iSnapBar, iSnapButton[8], "r=r", MARGIN);
 }
 
 // --------------------------------------------------------------------
@@ -874,6 +982,10 @@ void AppUi::resetCombos()
 void AppUi::addComboColors(AttributeSeq &sym, AttributeSeq &abs)
 {
   iInUiUpdate = true;
+  [iSelector[EUiStroke] addItemWithTitle:C2N(IPEABSOLUTE)];
+  [iSelector[EUiFill] addItemWithTitle:C2N(IPEABSOLUTE)];
+  iComboContents[EUiStroke].push_back(IPEABSOLUTE);
+  iComboContents[EUiFill].push_back(IPEABSOLUTE);
   for (uint i = 0; i < sym.size(); ++i) {
     Color color = abs[i].color();
     NSImage *im = colorIcon(color, 12);
@@ -961,6 +1073,11 @@ BOOL AppUi::validateMenuItem(NSMenuItem *item, NSString *name)
       @"Hide Snap Toolbar";
     return YES;
   }
+  if ([name isEqualToString:@"propertiespanel"]) {
+    item.title = [iPropertiesBox isHidden] ? @"Show Properties Panel" :
+      @"Hide Properties Panel";
+    return YES;
+  }
   IpeAction *s = findAction(name);
   if (s) {
     if ([name isEqualToString:@"toggle_notes"])
@@ -1021,18 +1138,50 @@ void AppUi::fillDynamicSubmenu(NSMenuItem *item)
       ++count;
       [sm addItem:mi];
     }
+  } else if (item.tag == ESubmenuRecentFiles) {
+    for (auto & name : iRecentFiles) {
+      NSMenuItem *mi = [[IpeMenuItem alloc]
+			   initWithTitle:I2N(name)
+			       ipeAction:@"recent_file"
+			   keyEquivalent:@""];
+      [sm addItem:mi];
+    }
   }
 }
 
+void AppUi::setRecentFileMenu(const std::vector<String> & names)
+{
+  iRecentFiles.clear();
+  for (const auto & name : names)
+    iRecentFiles.push_back(name);
+}
+
 void AppUi::toggleSnapbarShown()
 {
   iSnapBar.hidden = ![iSnapBar isHidden];
   if ([iSnapBar isHidden]) {
-    activateConstraint(iViewToSnapBar, NO);
-    activateConstraint(iViewToTop, YES);
+    iViewToSnapBar.active = NO;
+    iViewToTop.active = YES;
   } else {
-    activateConstraint(iViewToTop, NO);
-    activateConstraint(iViewToSnapBar, YES);
+    iViewToTop.active = NO;
+    iViewToSnapBar.active = YES;
+  }
+  iCanvas->update();
+}
+
+void AppUi::togglePropertiesShown()
+{
+  auto hidden = ![iPropertiesBox isHidden];
+  iPropertiesBox.hidden = hidden;
+  iLayerBox.hidden = hidden;
+  if (hidden) {
+    iViewToProperties.active = NO;
+    iViewToLayers.active = NO;
+    iViewToLeft.active = YES;
+  } else {
+    iViewToProperties.active = YES;
+    iViewToLayers.active = YES;
+    iViewToLeft.active = NO;
   }
   iCanvas->update();
 }
@@ -1145,7 +1294,7 @@ IpeAction *AppUi::findAction(NSString *name) const
 }
 
 // Determine if an action is checked.
-// Used for viewmarked, pagemarked, snapXXX, grid_visible,
+// Used for viewmarked, pagemarked, snapXXX, grid_visible, show_axes,
 // pretty_display, toggle_notes, toggle_bookmarks.
 bool AppUi::actionState(const char *name)
 {
@@ -1173,11 +1322,29 @@ void AppUi::setActionState(const char *name, bool value)
       iModeIndicator.image = loadIcon(name);
       iWindow.toolbar.selectedItemIdentifier = a;
     }
-  } else if ([a hasPrefix:@"snap"] || [a isEqualToString:@"grid_visible"]) {
-    for (int i = 0; i < 8; ++i) {
+  } else if ([a hasPrefix:@"snap"] || [a isEqualToString:@"grid_visible"] ||
+	     [a isEqualToString:@"show_axes"]) {
+    for (int i = 0; i < NUMSNAPBUTTONS; ++i) {
       if ([a isEqualToString:C2N(snapbutton_action[i])])
 	iSnapButton[i].state = value ? NSOnState : NSOffState;
     }
+    // update touch bar if it exists
+#if MAC_OS_X_VERSION_MAX_ALLOWED >= 101202
+#pragma clang diagnostic push
+#pragma clang diagnostic ignored "-Wdeprecated-declarations"
+    if ([iDelegate respondsToSelector:@selector(touchBar)]) {
+      id tbar = [iDelegate performSelector:@selector(touchBar)];
+      if (tbar != nil) {
+	auto bar = (NSTouchBar *) tbar;
+	auto item = [bar itemForIdentifier:a];
+	if (item != nil) {
+	  auto b = (NSButton *) item.view;
+	  b.state =  value ? NSOnState : NSOffState;
+	}
+      }
+    }
+#pragma clang diagnostic push
+#endif
   } else if ([a isEqualToString:@"shift_key"]) {
     int mod = s.state ? CanvasBase::EShift : 0;
     if (iCanvas)
@@ -1187,7 +1354,12 @@ void AppUi::setActionState(const char *name, bool value)
 
 void AppUi::action(String name)
 {
-  ipeDebug("action %s", name.z());
+  // ipeDebug("action %s", name.z());
+  if (name == "escape") {
+    if (iCanvas && iCanvas->tool())
+      iCanvas->tool()->key("\033", 0);
+    return;
+  }
   // Implement radio buttons
   int i = name.find("|");
   if (i >= 0)
@@ -1215,18 +1387,23 @@ void AppUi::setWindowCaption(bool mod, const char *s)
   [iWindow setTitle:[[NSString alloc] initWithUTF8String:s]];
 }
 
-void AppUi::showWindow(int width, int height)
+void AppUi::showWindow(int width, int height, int x, int y)
 {
   if (width > 0 && height > 0) {
     NSRect e = [[NSScreen mainScreen] frame];
     auto wd = e.size.width - width;
     auto hd = e.size.height - height;
-    NSRect winr = NSMakeRect(0.5 * wd, 0.5 * hd, width, height);
+    NSRect winr = NSMakeRect(x < 0 ? 0.5 * wd : x, y < 0 ? 0.5 * hd : y, width, height);
     [iWindow setFrame:winr display:YES];
   }
   [iWindow makeKeyAndOrderFront:iWindow];
 }
 
+void AppUi::setFullScreen(int mode)
+{
+  // not implemented, as OSX only provides toggleFullScreen
+}
+
 // show for t milliseconds, or permanently if t == 0
 void AppUi::explain(const char *s, int t)
 {
@@ -1292,12 +1469,73 @@ void AppUi::closeWindow()
 
 // --------------------------------------------------------------------
 
+NSImage *AppUi::createIcon(int pno, int size, bool touchBar)
+{
+  NSSize s = NSMakeSize(size, size);
+  int w = 2 * size;  // icons are square
+  int h = 2 * size;
+  return [NSImage imageWithSize:s
+			flipped:YES
+		 drawingHandler:^(NSRect rect) {
+      auto col = [[NSColor textColor] colorUsingColorSpace:[NSColorSpace genericGrayColorSpace]];
+      bool dark = touchBar || (col.whiteComponent > 0.5);
+      const Document *doc = dark ? ipeIconsDark.get() : ipeIcons.get();
+      Thumbnail thumbs(doc, w);
+      thumbs.setTransparent(true);
+      Buffer b = thumbs.render(doc->page(pno), 0);
+      cairo_surface_t *image =
+	cairo_image_surface_create_for_data((uint8_t *) b.data(),
+					    CAIRO_FORMAT_ARGB32,
+					    w, h, 4 * w);
+      CGContextRef ctx = [[NSGraphicsContext currentContext] CGContext];
+      cairo_surface_t *surface =
+	cairo_quartz_surface_create_for_cg_context(ctx, w, h);
+      cairo_t *cr = cairo_create(surface);
+      cairo_set_source_surface(cr, image, 0.0, 0.0);
+      cairo_matrix_t matrix;
+      cairo_matrix_init_scale(&matrix, 2.0, 2.0);
+      cairo_pattern_set_matrix(cairo_get_source(cr), &matrix);
+      cairo_paint(cr);
+      cairo_destroy(cr);
+      cairo_surface_finish(surface);
+      cairo_surface_destroy(surface);
+      cairo_surface_destroy(image);
+      return YES;
+    }];
+}
+
+NSImage *AppUi::loadIcon(String action, bool touchBar, int size)
+{
+  int pno = ipeIcon(action);
+  if (pno >= 0)
+    return createIcon(pno, size, touchBar);
+  else
+    // fallback if no icon has been defined
+    return colorIcon(0.8, 0.5, 0.7, size-2);
+}
+
+// --------------------------------------------------------------------
+
 int AppUi::clipboard(lua_State *L)
 {
-  // bool allow_bitmap = lua_toboolean(L, 2);
-  // NSImage can be read in the same way, but I don't
-  // know how to convert it to a bitmap...
   NSPasteboard *pasteBoard = [NSPasteboard generalPasteboard];
+  bool allow_bitmap = lua_toboolean(L, 2);
+  if (allow_bitmap) {
+    NSBitmapImageRep *rep = (NSBitmapImageRep *)[NSBitmapImageRep imageRepWithPasteboard:pasteBoard];
+    if (rep) {
+      CGImageRef bitmap = [rep CGImage];
+      int w = CGImageGetWidth(bitmap);
+      int h = CGImageGetHeight(bitmap);
+      Buffer data(w * h * sizeof(uint32_t));
+      if (cgImageDecode(bitmap, data)) {
+	Bitmap bitmap(w, h, Bitmap::ENative, data);
+	ipe::Rect r(Vector::ZERO, Vector(w, h));
+	Image *im = new Image(r, bitmap);
+	push_object(L, im);
+	return 1;
+      }
+    }
+  }
   NSArray *arr = [pasteBoard readObjectsForClasses:@[[NSString class]]
 					   options:@{ } ];
   if (arr && [arr count] > 0) {
