@@ -4,7 +4,7 @@
 /*
 
     This file is part of the extensible drawing editor Ipe.
-    Copyright (c) 1993-2019 Otfried Cheong
+    Copyright (c) 1993-2020 Otfried Cheong
 
     Ipe is free software; you can redistribute it and/or modify it
     under the terms of the GNU General Public License as published by
@@ -68,9 +68,9 @@ CanvasBase::CanvasBase()
 {
   iObserver = nullptr;
   iTool = nullptr;
-  iPage = 0;
-  iCascade = 0;
-  iSurface = 0;
+  iPage = nullptr;
+  iCascade = nullptr;
+  iSurface = nullptr;
   iPan = Vector::ZERO;
   iZoom = 1.0;
   iDimmed = false;
@@ -79,13 +79,14 @@ CanvasBase::CanvasBase()
   iBWidth = 0;
   iBHeight = 0;
   iRepaintObjects = false;
-  iFonts = nullptr;
   iResources = nullptr;
   iAutoSnap = false;
   iFifiVisible = false;
   iFifiMode = Snap::ESnapNone;
   iSelectionVisible = true;
 
+  iType3Font = false;
+
   isInkMode = false;
   iAdditionalModifiers = 0;
 
@@ -116,7 +117,6 @@ CanvasBase::~CanvasBase()
 {
   if (iSurface)
     cairo_surface_destroy(iSurface);
-  delete iFonts;
   delete iTool;
   ipeDebug("CanvasBase::~CanvasBase");
 }
@@ -126,9 +126,16 @@ CanvasBase::~CanvasBase()
 //! set information about Latex fonts (from ipe::Document)
 void CanvasBase::setResources(const PdfResources *resources)
 {
-  delete iFonts;
+  iFonts.reset();
   iResources = resources;
-  iFonts = Fonts::New(resources);
+  iFonts = std::make_unique<Fonts>(resources);
+}
+
+bool CanvasBase::type3Font()
+{
+  bool result = iType3Font;
+  iType3Font = false;
+  return result;
 }
 
 // --------------------------------------------------------------------
@@ -358,9 +365,10 @@ void CanvasBase::drawObjects(cairo_t *cc)
     cairo_clip(cc);
   }
 
-  CairoPainter painter(iCascade, iFonts, cc, iZoom, iStyle.pretty);
+  CairoPainter painter(iCascade, iFonts.get(), cc, iZoom, iStyle.pretty);
   painter.setDimmed(iDimmed);
-  // painter.Transform(CanvasTfm());
+  painter.setAttributeMap(&iPage->viewMap(iView));
+  std::vector<Matrix> layerMatrices = iPage->layerMatrices(iView);
   painter.pushMatrix();
 
   const Symbol *background =
@@ -379,10 +387,16 @@ void CanvasBase::drawObjects(cairo_t *cc)
     title->draw(painter);
 
   for (int i = 0; i < iPage->count(); ++i) {
-    if (iPage->objectVisible(iView, i))
+    if (iPage->objectVisible(iView, i)) {
+      painter.pushMatrix();
+      painter.transform(layerMatrices[iPage->layerOf(i)]);
       iPage->object(i)->draw(painter);
+      painter.popMatrix();
+    }
   }
   painter.popMatrix();
+  if (painter.type3Font())
+    iType3Font = true;
 }
 
 // --------------------------------------------------------------------
@@ -522,7 +536,7 @@ void CanvasBase::setTool(Tool *tool)
 void CanvasBase::finishTool()
 {
   delete iTool;
-  iTool = 0;
+  iTool = nullptr;
   iDimmed = false;
   iAutoSnap = false;
   update();
@@ -637,7 +651,7 @@ void CanvasBase::computeFifi(double x, double y)
   if (iSnap.iSnap & ~mask) {
     iFifiMode = iSnap.snap(iMousePos, iPage, iView,
 			   iSnap.iSnapDistance / iZoom,
-			   iTool, (iAutoSnap ? &iAutoOrigin : 0));
+			   iTool, (iAutoSnap ? &iAutoOrigin : nullptr));
     if (iFifiMode == Snap::ESnapNone)
       snapToPaperAndFrame();
 
@@ -665,8 +679,8 @@ void CanvasBase::refreshSurface()
       || iBWidth != cairo_image_surface_get_width(iSurface)
       || iBHeight != cairo_image_surface_get_height(iSurface)) {
     // size has changed
-    // ipeDebug("size has changed to %g x %g (%g x %g)",
-    //          iWidth, iHeight, iBWidth, iBHeight);
+    ipeDebug("size has changed to %g x %g (%g x %g)",
+	     iWidth, iHeight, iBWidth, iBHeight);
     if (iSurface)
       cairo_surface_destroy(iSurface);
     iSurface = nullptr;
@@ -678,8 +692,7 @@ void CanvasBase::refreshSurface()
   if (iRepaintObjects) {
     iRepaintObjects = false;
     if (!iSurface)
-      iSurface = cairo_image_surface_create(CAIRO_FORMAT_RGB24,
-					    iBWidth, iBHeight);
+      iSurface = cairo_image_surface_create(CAIRO_FORMAT_RGB24, iBWidth, iBHeight);
     cairo_t *cc = cairo_create(iSurface);
     // background
     cairo_set_source_rgb(cc, 0.4, 0.4, 0.4);
