@@ -4,7 +4,7 @@
 /*
 
     This file is part of the extensible drawing editor Ipe.
-    Copyright (c) 1993-2019 Otfried Cheong
+    Copyright (c) 1993-2020 Otfried Cheong
 
     Ipe is free software; you can redistribute it and/or modify it
     under the terms of the GNU General Public License as published by
@@ -176,9 +176,13 @@ bool ImlParser::parseBitmap()
     return false;
   String objNumStr;
   if (att.slash() && att.has("pdfObject", objNumStr)) {
-    int objNum = Lex(objNumStr).getInt();
-    // ipeDebug("ParseBitmap: objNum = %d", objNum);
-    Bitmap bitmap(att, pdfStream(objNum));
+    Lex lex(objNumStr);
+    Buffer data = pdfStream(lex.getInt());
+    Buffer alpha;
+    lex.skipWhitespace();
+    if (!lex.eos())
+      alpha = pdfStream(lex.getInt());
+    Bitmap bitmap(att, data, alpha);
     iBitmaps.push_back(bitmap);
   } else {
     String bits;
@@ -235,6 +239,16 @@ bool ImlParser::parsePage(Page &page)
     page.addLayer(att["name"]);
     if (att["edit"] == "no")
       page.setLocked(page.countLayers() - 1, true);
+    String snapMode;
+    if (att.has("snap", snapMode)) {
+      if (snapMode == "never")
+	page.setSnapping(page.countLayers() - 1, Page::SnapMode::Never);
+      else if (snapMode == "always")
+	page.setSnapping(page.countLayers() - 1, Page::SnapMode::Always);
+    }
+    String data;
+    if (att.has("data", data))
+      page.setLayerData(page.countLayers() - 1, data);
     tag = parseToTag();
   }
   // default layer: 'alpha'
@@ -270,6 +284,15 @@ bool ImlParser::parsePage(Page &page)
     if (att["marked"] == "yes")
       page.setMarkedView(page.countViews() - 1, true);
 
+    if (att.has("name", str))
+      page.setViewName(page.countViews() - 1, str);
+
+    if (!att.slash()) {
+      AttributeMap map;
+      if (!parseView(page, map))
+	return false;
+      page.setViewMap(page.countViews() - 1, map);
+    }
     tag = parseToTag();
   }
 
@@ -307,77 +330,39 @@ Page *ImlParser::parsePageSelection()
 {
   String tag = parseToTag();
   if (tag != "ipepage")
-    return 0;
+    return nullptr;
   XmlAttributes attr;
   if (!parseAttributes(attr))
-    return 0;
+    return nullptr;
   tag = parseToTag();
 
   while (tag == "bitmap") {
     if (!parseBitmap())
-      return 0;
+      return nullptr;
     tag = parseToTag();
   }
 
   if (tag != "page")
-    return 0;
+    return nullptr;
 
   std::unique_ptr<Page> page(new Page);
   if (!parsePage(*page))
-    return 0;
+    return nullptr;
 
   tag = parseToTag();
   if (tag != "/ipepage")
-    return 0;
+    return nullptr;
   return page.release();
 }
 
-#if 0
-//! parse an \c <ipeselection> element.
-/*! An PgObjectSeq is used to own the objects, but selection mode
-  and layer are not set. */
-bool ImlParser::parseSelection(PageObjectSeq &seq)
-{
-  String tag = parseToTag();
-  if (tag != "ipeselection")
-    return false;
-  XmlAttributes attr;
-  if (!parseAttributes(attr))
-    return 0;
-  tag = parseToTag();
-
-  while (tag == "bitmap") {
-    if (!parseBitmap())
-      return false;
-    tag = parseToTag();
-  }
-
-  for (;;) {
-    if (tag == "/ipeselection")
-      return true;
-    Object *obj = parseObject(tag);
-    if (!obj)
-      return false;
-    seq.push_back(PageObject(ENotSelected, 0, obj));
-    tag = parseToTag();
-  }
-}
-#endif
-
 //! parse an Object.
 /*! On calling, stream must be just past the tag. */
 Object *ImlParser::parseObject(String tag, Page *page,
 			       int *currentLayer)
 {
-  if (tag[0] == '/')
-    return 0;
-
-  XmlAttributes attr;
-  if (!parseAttributes(attr))
-    return 0;
-
   String layer;
-  if (page && currentLayer && attr.has("layer", layer)) {
+  Object *obj = parseObject(tag, layer);
+  if (obj && page && currentLayer && !layer.empty()) {
     for (int i = 0; i < page->countLayers(); ++i) {
       if (page->layer(i) == layer) {
 	*currentLayer = i;
@@ -385,6 +370,23 @@ Object *ImlParser::parseObject(String tag, Page *page,
       }
     }
   }
+  return obj;
+}
+
+//! parse an Object.
+/*! On calling, stream must be just past the tag. */
+Object *ImlParser::parseObject(String tag, String &layer)
+{
+  if (tag[0] == '/')
+    return nullptr;
+
+  XmlAttributes attr;
+  if (!parseAttributes(attr))
+    return nullptr;
+
+  String l;
+  if (attr.has("layer", l))
+    layer = l;
 
   if (tag == "group") {
     Group group(attr);
@@ -395,14 +397,14 @@ Object *ImlParser::parseObject(String tag, Page *page,
       }
       Object *obj = parseObject(tag);
       if (!obj)
-	return 0;
+	return nullptr;
       group.push_back(obj);
     }
   }
 
   String pcdata;
   if (!attr.slash() && !parsePCDATA(tag, pcdata))
-    return 0;
+    return nullptr;
   String bitmapId;
   if (tag == "image" && attr.has("bitmap", bitmapId)) {
     int objNum = Lex(bitmapId).getInt();
@@ -426,6 +428,61 @@ static inline bool symbolName(String s)
 			 ('A' <= s[0] && s[0] <= 'Z')));
 }
 
+//! Parse an attribute map.
+/*! On calling, stream must be before the first mapping element. */
+bool ImlParser::parseView(Page &page, AttributeMap &map)
+{
+  String tag = parseToTag();
+  while (tag != "/view") {
+    if (tag == "transform") {
+      XmlAttributes att;
+      if (!parseAttributes(att) || !att.slash())
+	return false;
+      String layer = att["layer"];
+      String mstr;
+      if (layer.empty() || !att.has("matrix", mstr))
+	return false;
+      int layerNum = page.findLayer(layer);
+      if (layerNum < 0)
+	return false;
+      page.setLayerMatrix(page.countViews() - 1, layerNum, Matrix(mstr));
+    } else {
+      XmlAttributes att;
+      if (!parseAttributes(att) || !att.slash())
+	return false;
+      String from = att["from"];
+      String kindStr = att["kind"];
+      Kind kind;
+      if (tag != "map") {  // deprecated format (briefly allowed in 7.2.16)
+	kindStr = tag;
+	from = att["map"];
+      }
+      if (kindStr == "pen")
+	kind = EPen;
+      else if (kindStr == "symbolsize")
+	kind = ESymbolSize;
+      else if (kindStr == "arrowsize")
+	kind = EArrowSize;
+      else if (kindStr == "opacity")
+	kind = EOpacity;
+      else if (kindStr == "color")
+	kind = EColor;
+      else if (kindStr == "dashstyle")
+	kind = EDashStyle;
+      else if (kindStr == "symbol")
+	kind = ESymbol;
+      else
+	return false; // error
+      String to = att["to"];
+      if (from.empty() || to.empty())
+	return false;
+      map.add({ kind, Attribute(true, from), Attribute(true, to) });
+    }
+    tag = parseToTag();
+  }
+  return true;
+}
+
 //! Parse a style sheet.
 /*! On calling, stream must be just past the style tag. */
 bool ImlParser::parseStyle(StyleSheet &sheet)
@@ -461,7 +518,7 @@ bool ImlParser::parseStyle(StyleSheet &sheet)
       else if (att["transformations"] == "translations")
 	symbol.iTransformations = ETransformationsTranslations;
       if (att["xform"] == "yes") {
-	uint flags = Reference::flagsFromName(name);
+	uint32_t flags = Reference::flagsFromName(name);
 	if ((flags & (Reference::EHasStroke|
 		      Reference::EHasFill|
 		      Reference::EHasPen|
@@ -470,6 +527,13 @@ bool ImlParser::parseStyle(StyleSheet &sheet)
 	  symbol.iTransformations = ETransformationsTranslations;
 	}
       }
+      Lex snaplex(att["snap"]);
+      while (!snaplex.eos()) {
+	Vector pos;
+	snaplex >> pos.x >> pos.y;
+	snaplex.skipWhitespace();
+	symbol.iSnap.push_back(pos);
+      }
       sheet.addSymbol(Attribute(true, name), symbol);
       if (parseToTag() != "/symbol")
 	return false;
@@ -618,7 +682,7 @@ bool ImlParser::parseStyle(StyleSheet &sheet)
       }
       if (s.iStops.front().offset < 0.0 || s.iStops.front().offset > 1.0)
 	return false;
-      for (uint i = 1; i < s.iStops.size(); ++i) {
+      for (int i = 1; i < size(s.iStops); ++i) {
 	if (s.iStops[i].offset < s.iStops[i-1].offset)
 	  return false;
       }
@@ -705,16 +769,16 @@ StyleSheet *ImlParser::parseStyleSheet()
   if (tag == "?xml") {
     XmlAttributes attr;
     if (!parseAttributes(attr, true))
-      return 0;
+      return nullptr;
     tag = parseToTag();
   }
   if (tag != "ipestyle")
-    return 0;
+    return nullptr;
   StyleSheet *sheet = new StyleSheet();
   if (parseStyle(*sheet))
     return sheet;
   delete sheet;
-  return 0;
+  return nullptr;
 }
 
 // --------------------------------------------------------------------
