@@ -4,7 +4,7 @@
 /*
 
     This file is part of the extensible drawing editor Ipe.
-    Copyright (c) 1993-2019 Otfried Cheong
+    Copyright (c) 1993-2020 Otfried Cheong
 
     Ipe is free software; you can redistribute it and/or modify it
     under the terms of the GNU General Public License as published by
@@ -42,24 +42,11 @@
 
 // --------------------------------------------------------------------
 
-void WString::init(const char *s)
+void WString::init(const char *s, int len)
 {
-  int rw = MultiByteToWideChar(CP_UTF8, 0, s, -1, NULL, 0);
-  iData.resize(rw);
-  MultiByteToWideChar(CP_UTF8, 0, s, -1, &iData[0], rw);
-}
-
-void WString::appendTo(std::vector<wchar_t> &out) const
-{
-  for (int i = 0; i < int(iData.size()); ++i)
-    out.push_back(iData[i]);
-}
-
-void WString::append(const WString &s)
-{
-  iData.pop_back();
-  iData.reserve(iData.size() + s.size());
-  s.appendTo(iData);
+  int rw = MultiByteToWideChar(CP_UTF8, 0, s, len, nullptr, 0);
+  resize(rw + 1, wchar_t(0));
+  MultiByteToWideChar(CP_UTF8, 0, s, len, data(), rw);
 }
 
 BOOL setWindowText(HWND h, const char *s)
@@ -88,9 +75,9 @@ void sendMessage(HWND h, UINT code, const char *t, WPARAM wParam)
 
 static std::string wideToUtf8(const wchar_t *wbuf)
 {
-  int rm = WideCharToMultiByte(CP_UTF8, 0, wbuf, -1, NULL, 0, NULL, NULL);
+  int rm = WideCharToMultiByte(CP_UTF8, 0, wbuf, -1, nullptr, 0, nullptr, nullptr);
   std::vector<char> multi(rm);
-  WideCharToMultiByte(CP_UTF8, 0, wbuf, -1, &multi[0], rm, NULL, NULL);
+  WideCharToMultiByte(CP_UTF8, 0, wbuf, -1, &multi[0], rm, nullptr, nullptr);
   return std::string(&multi[0]);
 }
 
@@ -110,13 +97,13 @@ static void buildFlags(std::vector<short> &t, DWORD flags)
 static void buildString(std::vector<short> &t, const char *s)
 {
   WString w(s);
-  const wchar_t *p = w;
+  const wchar_t *p = w.data();
   while (*p)
     t.push_back(*p++);
   t.push_back(0);
 }
 
-static void buildControl(std::vector<short> &t, short what, const char *s=0)
+static void buildControl(std::vector<short> &t, short what, const char *s = nullptr)
 {
   t.push_back(0xffff);
   t.push_back(what);
@@ -142,12 +129,14 @@ protected:
   virtual void enableItem(int idx, bool value);
 
   void computeDimensions(int &w, int &h);
+  void getDimensions(SElement &m, int &x, int &y, int &w, int &h);
   void buildDimensions(std::vector<short> &t, SElement &m, int id);
 
   virtual void acceptDialog(lua_State *L);
 
   BOOL initDialog();
   BOOL dlgCommand(WPARAM wParam, LPARAM lParam);
+  BOOL handleResize();
   static BOOL CALLBACK dialogProc(HWND hwndDlg, UINT message,
 				  WPARAM wParam, LPARAM lParam);
 private:
@@ -319,12 +308,12 @@ void PDialog::retrieveValues()
   }
 }
 
-void PDialog::buildDimensions(std::vector<short> &t, SElement &m, int id)
+void PDialog::getDimensions(SElement &m, int &x, int &y, int &w, int &h)
 {
-  int x = BORDER;
-  int y = BORDER;
-  int w = 0;
-  int h = 0;
+  x = BORDER;
+  y = BORDER;
+  w = 0;
+  h = 0;
   if (m.row < 0) {
     y = BORDER;
     for (int i = 0; i < iNoRows; ++i)
@@ -345,6 +334,12 @@ void PDialog::buildDimensions(std::vector<short> &t, SElement &m, int id)
     for (int i = m.row + 1; i < m.row + m.rowspan; ++i)
       h += iRowHeight[i] + PAD;
   }
+}
+
+void PDialog::buildDimensions(std::vector<short> &t, SElement &m, int id)
+{
+  int x, y, w, h;
+  getDimensions(m, x, y, w, h);
   t.push_back(x);
   t.push_back(y);
   t.push_back(w);
@@ -360,10 +355,14 @@ BOOL PDialog::dlgCommand(WPARAM wParam, LPARAM lParam)
     EndDialog(hDialog, TRUE);
     return TRUE;
   }
-  if (id == IDCANCEL && !iIgnoreEscape) {
+  if (id == IDCANCEL) {
     retrieveValues();
-    EndDialog(hDialog, FALSE);
-    return TRUE;
+    if (iIgnoreEscapeField < 0 ||
+	iElements[iIgnoreEscapeField].text == iIgnoreEscapeText) {
+      // allowed to close dialog
+      EndDialog(hDialog, FALSE);
+      return TRUE;
+    }
   }
   if (id < IDBASE || id >= IDBASE + int(iElements.size()))
     return FALSE;
@@ -427,6 +426,10 @@ BOOL CALLBACK PDialog::dialogProc(HWND hwnd, UINT message,
       return d->dlgCommand(wParam, lParam);
     else
       return FALSE;
+  case WM_SIZE:
+    if (d)
+      return d->handleResize();
+    return FALSE;
   case WM_DESTROY:
     // Remove the subclasses from text edits
     for (int i = 0; i < int(d->iElements.size()); ++i) {
@@ -501,6 +504,7 @@ void PDialog::buildElements(std::vector<short> &t)
 
 bool PDialog::buildAndRun(int w, int h)
 {
+  // converts dimensions to dialog units and possibly changes them
   computeDimensions(w, h);
 
   RECT rect;
@@ -511,12 +515,12 @@ bool PDialog::buildAndRun(int w, int h)
   std::vector<short> t;
 
   // Dialog flags
-  buildFlags(t, WS_POPUP | WS_BORDER | DS_SHELLFONT |
+  buildFlags(t, WS_POPUP | WS_BORDER | DS_SHELLFONT | WS_SIZEBOX |
 	     WS_SYSMENU | DS_MODALFRAME | WS_CAPTION);
   t.push_back(iElements.size());
-  t.push_back((pw - w)/2);  // offset of popup-window from parent window
+  t.push_back((pw - w)/2);  // offset of popup-window from parent window in dialog units
   t.push_back((ph - h)/2);
-  t.push_back(w);
+  t.push_back(w);  // should be dialog units
   t.push_back(h);
   // menu
   t.push_back(0);
@@ -534,10 +538,11 @@ bool PDialog::buildAndRun(int w, int h)
 			    (LPCDLGTEMPLATE) &t[0],
 			    iParent, (DLGPROC) dialogProc, (LPARAM) this);
   // retrieveValues() has been called before EndDialog!
-  hDialog = NULL; // already destroyed by Windows
+  hDialog = nullptr; // already destroyed by Windows
   return (res > 0);
 }
 
+// converts (w, h) from pixels to dialog units and possibly increases them to fit dialog
 void PDialog::computeDimensions(int &w, int &h)
 {
   int minWidth[iNoCols];
@@ -626,8 +631,9 @@ void PDialog::computeDimensions(int &w, int &h)
       iRowHeight[i] = minHeight[i] + extra;
     }
   }
+
   /*
-  fprintf(stderr, "iColWidth\n");
+  fprintf(stderr, "iNoCols: %d, iColWidth = ", iNoCols);
   for (int i = 0; i < iNoCols; ++i)
     fprintf(stderr, "%d ", iColWidth[i]);
   fprintf(stderr, "\n");
@@ -635,6 +641,27 @@ void PDialog::computeDimensions(int &w, int &h)
   */
 }
 
+BOOL PDialog::handleResize()
+{
+  RECT rc;
+  GetClientRect(hDialog, &rc);
+  int bw = rc.right - rc.left;
+  int bh = rc.bottom - rc.top;
+  computeDimensions(bw, bh);
+
+  for (int i = 0; i < int(iElements.size()); ++i) {
+    SElement &m = iElements[i];
+    int id = i + IDBASE;
+    int x, y, w, h;
+    getDimensions(m, x, y, w, h);
+    HWND hwnd = GetDlgItem(hDialog, id);
+    // calculation is in dialog units, so convert back to pixels
+    MoveWindow(hwnd, x * iBaseX / 4 , y * iBaseY / 8 ,
+	       w * iBaseX / 4, h * iBaseY / 8, TRUE);
+  }
+  return TRUE;
+}
+
 void PDialog::enableItem(int idx, bool value)
 {
   EnableWindow(GetDlgItem(hDialog, idx+IDBASE), value);
@@ -645,10 +672,10 @@ void PDialog::enableItem(int idx, bool value)
 static int dialog_constructor(lua_State *L)
 {
   HWND parent = check_winid(L, 1);
-  const char *s = luaL_checkstring(L, 2);
+  const char *s = luaL_checklstring(L, 2, nullptr);
 
   Dialog **dlg = (Dialog **) lua_newuserdata(L, sizeof(Dialog *));
-  *dlg = 0;
+  *dlg = nullptr;
   luaL_getmetatable(L, "Ipe.dialog");
   lua_setmetatable(L, -2);
   *dlg = new PDialog(L, parent, s);
@@ -686,7 +713,7 @@ PMenu::~PMenu()
 {
   if (hMenu)
     DestroyMenu(hMenu);
-  hMenu = 0;
+  hMenu = nullptr;
   for (int i = 0; i < int(bitmaps.size()); ++i)
     DeleteObject(bitmaps[i]);
 }
@@ -697,7 +724,7 @@ int PMenu::execute(lua_State *L)
   int vy = (int)luaL_checkinteger(L, 3);
   int result = TrackPopupMenu(hMenu,
 			      TPM_NONOTIFY | TPM_RETURNCMD | TPM_RIGHTBUTTON,
-			      vx, vy, 0, hwnd, NULL);
+			      vx, vy, 0, hwnd, nullptr);
   if (1 <= result && result <= int(items.size())) {
     result -= 1;
     lua_pushstring(L, items[result].name.c_str());
@@ -719,7 +746,7 @@ static HBITMAP colorIcon(double red, double green, double blue)
   COLORREF rgb = RGB(r, g, b);
   int cx = GetSystemMetrics(SM_CXMENUCHECK);
   int cy = GetSystemMetrics(SM_CYMENUCHECK);
-  HDC hdc = GetDC(NULL);
+  HDC hdc = GetDC(nullptr);
   HDC memDC = CreateCompatibleDC(hdc);
   HBITMAP bm = CreateCompatibleBitmap(hdc, cx, cy);
   SelectObject(memDC, bm);
@@ -729,16 +756,16 @@ static HBITMAP colorIcon(double red, double green, double blue)
     }
   }
   DeleteDC(memDC);
-  ReleaseDC(NULL, hdc);
+  ReleaseDC(nullptr, hdc);
   return bm;
 }
 
 int PMenu::add(lua_State *L)
 {
-  const char *name = luaL_checkstring(L, 2);
-  WString title(luaL_checkstring(L, 3));
+  const char *name = luaL_checklstring(L, 2, nullptr);
+  WString title(luaL_checklstring(L, 3, nullptr));
   if (lua_gettop(L) == 3) {
-    AppendMenuW(hMenu, MF_STRING, items.size() + 1, title);
+    AppendMenuW(hMenu, MF_STRING, items.size() + 1, title.data());
     Item item;
     item.name = name;
     item.itemIndex = 0;
@@ -752,11 +779,11 @@ int PMenu::add(lua_State *L)
     if (hastable)
       luaL_argcheck(L, lua_istable(L, 5), 5,
 		    "argument is not a function or table");
-    const char *current = 0;
+    const char *current = nullptr;
     if (hascheck) {
       luaL_argcheck(L, lua_isstring(L, 6), 6,
 		    "argument is not a function or string");
-      current = luaL_checkstring(L, 6);
+      current = luaL_checklstring(L, 6, nullptr);
     }
     int no = lua_rawlen(L, 4);
     HMENU sm = CreatePopupMenu();
@@ -764,7 +791,7 @@ int PMenu::add(lua_State *L)
       lua_rawgeti(L, 4, i);
       luaL_argcheck(L, lua_isstring(L, -1), 4, "items must be strings");
       int id = items.size() + 1;
-      const char *item = lua_tostring(L, -1);
+      const char *item = lua_tolstring(L, -1, nullptr);
       if (hastable) {
 	lua_rawgeti(L, 5, i);
 	luaL_argcheck(L, lua_isstring(L, -1), 5, "labels must be strings");
@@ -772,14 +799,14 @@ int PMenu::add(lua_State *L)
 	lua_pushvalue(L, 5);   // function
 	lua_pushnumber(L, i);  // index
  	lua_pushvalue(L, -3);  // name
-	lua_call(L, 2, 1);     // function returns label
+	luacall(L, 2, 1);      // function returns label
 	luaL_argcheck(L, lua_isstring(L, -1), 5,
 		      "function does not return string");
       } else
 	lua_pushvalue(L, -1);
 
-      WString text(lua_tostring(L, -1));
-      AppendMenuW(sm, MF_STRING, id, text);
+      WString text(tostring(L, -1));
+      AppendMenuW(sm, MF_STRING, id, text.data());
       Item mitem;
       mitem.name = name;
       mitem.itemName = item;
@@ -792,7 +819,7 @@ int PMenu::add(lua_State *L)
 	lua_pushvalue(L, 6);   // function
 	lua_pushnumber(L, i);  // index
  	lua_pushvalue(L, -4);  // name
-	lua_call(L, 2, 3);     // function returns red, green, blue
+	luacall(L, 2, 3);      // function returns red, green, blue
 	double red = luaL_checknumber(L, -3);
 	double green = luaL_checknumber(L, -2);
 	double blue = luaL_checknumber(L, -1);
@@ -803,7 +830,7 @@ int PMenu::add(lua_State *L)
       }
       lua_pop(L, 2); // item, text
     }
-    AppendMenuW(hMenu, MF_STRING | MF_POPUP, (UINT_PTR) sm, title);
+    AppendMenuW(hMenu, MF_STRING | MF_POPUP, (UINT_PTR) sm, title.data());
   }
   return 0;
 }
@@ -814,7 +841,7 @@ static int menu_constructor(lua_State *L)
 {
   HWND hwnd = check_winid(L, 1);
   Menu **m = (Menu **) lua_newuserdata(L, sizeof(Menu *));
-  *m = 0;
+  *m = nullptr;
   luaL_getmetatable(L, "Ipe.menu");
   lua_setmetatable(L, -2);
   *m = new PMenu(hwnd);
@@ -867,7 +894,7 @@ PTimer::PTimer(lua_State *L0, int lua_object, const char *method)
 PTimer::~PTimer()
 {
   if (iTimer)
-    KillTimer(NULL, iTimer);
+    KillTimer(nullptr, iTimer);
   // remove it from all_timers
   for (int i = 0; i < int(all_timers.size()); ++i) {
     if (all_timers[i] == this) {
@@ -881,7 +908,7 @@ void PTimer::elapsed()
 {
   callLua();
   if (iSingleShot) {
-    KillTimer(NULL, iTimer);
+    KillTimer(nullptr, iTimer);
     iTimer = 0;
   }
 }
@@ -891,7 +918,7 @@ int PTimer::setInterval(lua_State *L)
   int t = (int)luaL_checkinteger(L, 2);
   iInterval = t;
   if (iTimer)
-    SetTimer(NULL, iTimer, iInterval, TimerProc);
+    SetTimer(nullptr, iTimer, iInterval, TimerProc);
   return 0;
 }
 
@@ -904,14 +931,14 @@ int PTimer::active(lua_State *L)
 int PTimer::start(lua_State *L)
 {
   if (iTimer == 0)
-    iTimer = SetTimer(NULL, 0, iInterval, TimerProc);
+    iTimer = SetTimer(nullptr, 0, iInterval, TimerProc);
   return 0;
 }
 
 int PTimer::stop(lua_State *L)
 {
   if (iTimer) {
-    KillTimer(NULL, iTimer);
+    KillTimer(nullptr, iTimer);
     iTimer = 0;
   }
   return 0;
@@ -922,10 +949,10 @@ int PTimer::stop(lua_State *L)
 static int timer_constructor(lua_State *L)
 {
   luaL_argcheck(L, lua_istable(L, 1), 1, "argument is not a table");
-  const char *method = luaL_checkstring(L, 2);
+  const char *method = luaL_checklstring(L, 2, nullptr);
 
   Timer **t = (Timer **) lua_newuserdata(L, sizeof(Timer *));
-  *t = 0;
+  *t = nullptr;
   luaL_getmetatable(L, "Ipe.timer");
   lua_setmetatable(L, -2);
 
@@ -979,29 +1006,29 @@ static int ipeui_getColor(lua_State *L)
 
 static int ipeui_fileDialog(lua_State *L)
 {
-  static const char * const typenames[] = { "open", "save", 0 };
+  static const char * const typenames[] = { "open", "save", nullptr };
   HWND hwnd = check_winid(L, 1);
-  int type = luaL_checkoption(L, 2, 0, typenames);
-  WString caption(luaL_checkstring(L, 3));
+  int type = luaL_checkoption(L, 2, nullptr, typenames);
+  WString caption(luaL_checklstring(L, 3, nullptr));
   if (!lua_istable(L, 4))
     luaL_argerror(L, 4, "table expected for filters");
-  std::vector<wchar_t> filters;
+  std::wstring filters;
   int nFilters = lua_rawlen(L, 4);
   for (int i = 1; i <= nFilters; ++i) {
     lua_rawgeti(L, 4, i);
     luaL_argcheck(L, lua_isstring(L, -1), 4, "filter entry is not a string");
-    WString el(lua_tostring(L, -1));
-    el.appendTo(filters);
+    WString el(tostring(L, -1));
+    filters += el;
     lua_pop(L, 1); // element i
   }
   filters.push_back(0); // terminate list
 
-  const char *dir = 0;
+  const char *dir = nullptr;
   if (!lua_isnoneornil(L, 5))
-    dir = luaL_checkstring(L, 5);
-  const char *name = 0;
+    dir = luaL_checklstring(L, 5, nullptr);
+  const char *name = nullptr;
   if (!lua_isnoneornil(L, 6))
-    name = luaL_checkstring(L, 6);
+    name = luaL_checklstring(L, 6, nullptr);
   int selected = 0;
   if (!lua_isnoneornil(L, 7))
     selected = luaL_checkinteger(L, 7);
@@ -1025,9 +1052,9 @@ static int ipeui_fileDialog(lua_State *L)
   }
   if (dir) {
     WString wdir(dir);
-    ofn.lpstrInitialDir = wdir;
+    ofn.lpstrInitialDir = wdir.data();
   }
-  ofn.lpstrTitle = caption;
+  ofn.lpstrTitle = caption.data();
 
   BOOL result;
   if (type == 0) {
@@ -1051,22 +1078,22 @@ static int ipeui_fileDialog(lua_State *L)
 static int ipeui_messageBox(lua_State *L)
 {
   static const char * const options[] =  {
-    "none", "warning", "information", "question", "critical", 0 };
+    "none", "warning", "information", "question", "critical", nullptr };
   static const char * const buttontype[] = {
     "ok", "okcancel", "yesnocancel", "discardcancel",
-    "savediscardcancel", 0 };
+    "savediscardcancel", nullptr };
 
   HWND hwnd = check_winid(L, 1);
   int type = luaL_checkoption(L, 2, "none", options);
-  const char *text = luaL_checkstring(L, 3);
-  const char *details = 0;
+  const char *text = luaL_checklstring(L, 3, nullptr);
+  const char *details = nullptr;
   if (!lua_isnoneornil(L, 4))
-    details = luaL_checkstring(L, 4);
+    details = luaL_checklstring(L, 4, nullptr);
   int buttons = 0;
   if (lua_isnumber(L, 5))
     buttons = (int)luaL_checkinteger(L, 5);
   else if (!lua_isnoneornil(L, 5))
-    buttons = luaL_checkoption(L, 5, 0, buttontype);
+    buttons = luaL_checkoption(L, 5, nullptr, buttontype);
 
   UINT uType = MB_APPLMODAL;
   switch (type) {
@@ -1113,11 +1140,9 @@ static int ipeui_messageBox(lua_State *L)
     char buf[strlen(text) + strlen(details) + 8];
     sprintf(buf, "%s\n\n%s", text, details);
     WString wbuf(buf);
-    ret = MessageBoxW(hwnd, wbuf, L"Ipe", uType);
-  } else {
-    WString wbuf(text);
-    ret = MessageBoxW(hwnd, wbuf, L"Ipe", uType);
-  }
+    ret = MessageBoxW(hwnd, wbuf.data(), L"Ipe", uType);
+  } else
+    ret = MessageBoxW(hwnd, WString(text).data(), L"Ipe", uType);
 
   switch (ret) {
   case IDOK:
@@ -1138,13 +1163,24 @@ static int ipeui_messageBox(lua_State *L)
 
 // --------------------------------------------------------------------
 
+struct SDialogHandle {
+  HWND hwnd;
+  HANDLE thread;
+};
+
 BOOL CALLBACK waitDialogProc(HWND hwnd, UINT message,
 			     WPARAM wParam, LPARAM lParam)
 {
   switch (message) {
   case WM_INITDIALOG: {
-    HWND *p = (HWND *) lParam;
-    *p = hwnd;
+    SDialogHandle *d = (SDialogHandle *) lParam;
+    d->hwnd = hwnd;
+    if (d->thread) {
+      ResumeThread(d->thread);
+      // delay showing the dialog by 300ms
+      // if Latex is fast, the dialog never gets shown
+      Sleep(300);
+    }
     return TRUE; }
   default:
     return FALSE;
@@ -1153,15 +1189,29 @@ BOOL CALLBACK waitDialogProc(HWND hwnd, UINT message,
 
 VOID CALLBACK waitCallback(PVOID lpParameter, BOOLEAN timerOrWaitFired)
 {
-  HWND *hDialog = (HWND *) lpParameter;
-  if (*hDialog)
-    EndDialog(*hDialog, 1);
+  SDialogHandle *d = (SDialogHandle *) lpParameter;
+  EndDialog(d->hwnd, 1);
+}
+
+DWORD CALLBACK waitLuaThreadProc(LPVOID lpParameter)
+{
+  lua_State *L = (lua_State *) lpParameter;
+  lua_callk(L, 0, 0, 0, nullptr);
+  return 0;
 }
 
 static int ipeui_wait(lua_State *L)
 {
-  Dialog **dlg = check_dialog(L, 1);
-  HWND parent = (*dlg)->winId();
+  Dialog **dlg = (Dialog **) luaL_testudata(L, 1, "Ipe.dialog");
+  HWND parent = dlg ? (*dlg)->winId() : check_winid(L, 1);
+
+  const char *cmd = nullptr;
+  if (!lua_isfunction(L, 2))
+    cmd = luaL_checklstring(L, 2, nullptr);
+
+  const char *label = "Waiting for external editor";
+  if (lua_isstring(L, 3))
+    label = luaL_checklstring(L, 3, nullptr);
 
   std::vector<short> t;
 
@@ -1190,44 +1240,61 @@ static int ipeui_wait(lua_State *L)
   t.push_back(120);
   t.push_back(20);
   t.push_back(IDBASE);
-  buildControl(t, 0x0082, "Waiting for external editor");
+  buildControl(t, 0x0082, label);
 
-  // Declare and initialize process blocks
-  PROCESS_INFORMATION processInformation;
-  STARTUPINFOW startupInfo;
+  SDialogHandle dialogHandle = { nullptr, nullptr };
 
-  memset(&processInformation, 0, sizeof(processInformation));
-  memset(&startupInfo, 0, sizeof(startupInfo));
-  startupInfo.cb = sizeof(startupInfo);
-
-  // Call the executable program
-  const char *cmd = luaL_checkstring(L, 2);
-  WString wcmd(cmd);
-
-  int result = CreateProcessW(NULL, wcmd, NULL, NULL, FALSE,
-			      NORMAL_PRIORITY_CLASS|CREATE_NO_WINDOW,
-			      NULL, NULL, &startupInfo, &processInformation);
-  if (result == 0)
-    return 0;
-
-  HWND dialogHandle = NULL;
   HANDLE waitHandle;
 
-  RegisterWaitForSingleObject(&waitHandle,
-			      processInformation.hProcess,
-			      waitCallback,
-			      &dialogHandle,
-			      INFINITE,
-			      WT_EXECUTEINWAITTHREAD|WT_EXECUTEONLYONCE);
+  if (cmd) {
+    // call external process
 
-  DialogBoxIndirectParamW((HINSTANCE) GetWindowLongPtr(parent, GWLP_HINSTANCE),
-			  (LPCDLGTEMPLATE) &t[0],
-			  parent, (DLGPROC) waitDialogProc,
-			  (LPARAM) &dialogHandle);
+    // Declare and initialize process blocks
+    PROCESS_INFORMATION processInformation;
+    STARTUPINFOW startupInfo;
+    memset(&processInformation, 0, sizeof(processInformation));
+    memset(&startupInfo, 0, sizeof(startupInfo));
+    startupInfo.cb = sizeof(startupInfo);
 
-  UnregisterWait(waitHandle);
-  CloseHandle(processInformation.hProcess);
-  CloseHandle(processInformation.hThread);
+    WString wcmd(cmd);
+
+    int result = CreateProcessW(nullptr, wcmd.data(), nullptr, nullptr, FALSE,
+				NORMAL_PRIORITY_CLASS|CREATE_NO_WINDOW,
+				nullptr, nullptr, &startupInfo, &processInformation);
+    if (result == 0)
+      return 0;
+
+    RegisterWaitForSingleObject(&waitHandle, processInformation.hProcess, waitCallback, &dialogHandle,
+				INFINITE, WT_EXECUTEINWAITTHREAD|WT_EXECUTEONLYONCE);
+
+    DialogBoxIndirectParamW((HINSTANCE) GetWindowLongPtr(parent, GWLP_HINSTANCE),
+			    (LPCDLGTEMPLATE) &t[0],
+			    parent, (DLGPROC) waitDialogProc,
+			    (LPARAM) &dialogHandle);
+
+    UnregisterWait(waitHandle);
+    CloseHandle(processInformation.hProcess);
+    CloseHandle(processInformation.hThread);
+
+  } else {
+    // call Lua function
+
+    dialogHandle.thread = CreateThread(nullptr, 0, &waitLuaThreadProc, L, CREATE_SUSPENDED, nullptr);
+    if (dialogHandle.thread == nullptr)
+      return 0;
+
+    RegisterWaitForSingleObject(&waitHandle, dialogHandle.thread, waitCallback, &dialogHandle,
+				INFINITE, WT_EXECUTEINWAITTHREAD|WT_EXECUTEONLYONCE);
+
+    lua_pushvalue(L, 2);
+
+    DialogBoxIndirectParamW((HINSTANCE) GetWindowLongPtr(parent, GWLP_HINSTANCE),
+			    (LPCDLGTEMPLATE) &t[0],
+			    parent, (DLGPROC) waitDialogProc,
+			    (LPARAM) &dialogHandle);
+    UnregisterWait(waitHandle);
+    CloseHandle(dialogHandle.thread);
+  }
   return 0;
 }
 
@@ -1247,8 +1314,8 @@ static int ipeui_currentDateTime(lua_State *L)
 
 static int ipeui_startBrowser(lua_State *L)
 {
-  const char *url = luaL_checkstring(L, 1);
-  long long int res = (long long int) ShellExecuteA(0, "open", url, 0, 0, SW_SHOWNORMAL);
+  const char *url = luaL_checklstring(L, 1, nullptr);
+  long long int res = (long long int) ShellExecuteA(nullptr, "open", url, nullptr, nullptr, SW_SHOWNORMAL);
   lua_pushboolean(L, (res >= 32));
   return 1;
 }
@@ -1265,7 +1332,7 @@ static const struct luaL_Reg ipeui_functions[] = {
   { "waitDialog", ipeui_wait },
   { "currentDateTime", ipeui_currentDateTime },
   { "startBrowser", ipeui_startBrowser },
-  { 0, 0},
+  { nullptr, nullptr }
 };
 
 // --------------------------------------------------------------------
